-- arch-tag: 17540d73-d26c-4eaf-a439-721b92c76b42

module Filter where

import SimpleParser
import Char
import List
import Puff
import PackedString
import Monad
import Regex

------------------------
-- Filter implementation
------------------------


data Filter = 
    FilterAuthor String 
    | FilterCategory Category  
    | FilterNot Filter 
    | FilterOr [Filter] 
    | FilterAnd [Filter] 
    | FilterPrivate 
    | FilterTrue 
    | FilterRegex PackedString Rx
    | FilterAny PackedString Rx


parseFilter :: Monad m => String -> m Filter
parseFilter = parser (between spaces eof rmp) where
    rmp = do
	fa <- mp 
	r <- rmp'
	case r of
	    [] -> return fa
	    _ -> return $ FilterOr (fa:r)
    rmp' = (do
	char ';'
	spaces
	v <- mp
	r <- rmp'
	return (v:r)) <|> return []
    mp = many1 fp >>= \v -> case v of
	    [x] -> return x
	    xs -> return $ FilterAnd xs
    fp = nf <|> flt <|> pr <|> bool <|> do s <- (token string); liftM (FilterRegex f_messageBody) (compileRx s) 
    pr = do
	char '('
	spaces
	v <- rmp
	char ')'
	spaces
	return v
    nf = char '!' >> liftM FilterNot fp
    string = qstring <|> many1 (noneOf " \t\n~'!();%")
    qstring = between (char '\'') (char '\'') $ many ((char '\'' >> char '\'' >> return '\'') <|> noneOf "'")
    bool = do
	char '%'
	c <- sat isAlpha
	spaces
	return $ b c
    flt = do
	char '~'
	c <- sat isAlpha
	option ':' (char ':')
	s <- string
	spaces
	z c s
    z 'a' s = return $ FilterAuthor s
    z 'c' s = return $ FilterCategory (catParseNew s)
    z 't' s = return $ FilterCategory (catParseNew s) -- TODO
    z 's' s = compileRx s >>= return . FilterRegex f_messageSender
    z 'k' s = compileRx s >>= return . FilterRegex f_messageKeyword
    z _ s = compileRx s >>= return . (FilterRegex f_messageBody)
    b 'p' = FilterPrivate
    b 'P' = FilterNot FilterPrivate
    b _ = FilterTrue

showFilter :: Filter -> String
showFilter (FilterAuthor a) = "~a:" ++  a
showFilter (FilterCategory c) = "~c:" ++ catShowNew c
showFilter (FilterAnd fs) = concat (intersperse " " (map showFilter fs))
showFilter (FilterOr fs) = "(" ++ (concat $ intersperse " ; " (map showFilter fs)) ++ ")"
showFilter FilterPrivate = "%p"
showFilter (FilterNot FilterPrivate) = "%P"
showFilter FilterTrue = "%t"
showFilter (FilterNot FilterTrue) = "%f"
showFilter (FilterRegex fn c) | fn == f_messageKeyword = "~k:" ++ show c
showFilter (FilterRegex fn s) | fn == f_messageSender = "~s:" ++ show s
showFilter (FilterRegex fn s) | fn == f_messageBody = show s 
showFilter (FilterAny fn r) = "~A:" ++ unpackPS fn ++ ":" ++ show r
showFilter (FilterRegex fn r) = "~R:" ++ unpackPS fn ++ ":" ++ show r
showFilter (FilterNot f) = "!" ++ showFilter f

applyFilterStack :: [Filter] -> [(Int, Puff)] -> [(Int, Puff)]
applyFilterStack fs = concatMap f where
    f a@(_,p) | all (flip filterAp p) fs = [a]
    f _ = []


siglookup c (Unverifyable (Key n _):_) | c == n = True
siglookup c (Signed (Key n _):_) | c == n = True
siglookup c (_:xs) = siglookup c xs
siglookup _ [] = False


filterAp :: Filter -> Puff -> Bool
filterAp (FilterNot f) p = not $ filterAp f p
filterAp (FilterAuthor c) (Puff {signature = sig})  = siglookup c sig 
filterAp (FilterCategory c) (Puff {cats = cs}) = any (`subCategory` c) cs
filterAp (FilterAnd fs) p = all (`filterAp` p) fs
filterAp (FilterOr fs) p = any (`filterAp` p) fs
filterAp (FilterAny fn re) p = any (matchRx re) (getFragmentForceStrings p fn)
filterAp (FilterRegex fn re) p = any (matchRx re) (getFragmentStrings p fn)
filterAp FilterTrue _ = True
filterAp FilterPrivate _ = False

--showFilter _ = ""


